#include "loadbspf.hpp"
#include "../common/bspfileformat.hpp"
#include "struct.hpp"
#include "utils.hpp"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

static bool loadlump(FILE *in,lump_t &lump)
{
  fread(&lump,sizeof(lump_t),1,in);
  if(lump.idlump== 'S'+256*'L') return true;
  else return false;
}

static bool loadvertex(FILE*in, map_t *m,lump_t lump)
{
  if(m->ve) {
    PRINTLOG(1,"more that one vertex lump");
    free(m->ve);
  }
  m->ve=(vertex_t*)malloc(sizeof(vertex_t)*lump.nbr);
  if(m->ve==NULL) {
    PRINTLOG(1,"no enugth memory");
    return false;
  }

  bspf_vertex_t *lo=(bspf_vertex_t*)malloc(lump.nbr*sizeof(bspf_vertex_t));
  fread(lo,sizeof(bspf_vertex_t),lump.nbr,in);

  //FIX
  for(int i=0;i<lump.nbr;i++) {
    //for(int j=0;j<3;j++) 
    //  lo[i].pos[j]/=1000.;
    for(int j=0;j<2;j++) 
      lo[i].texpos[j]/=100.;
    }

  for(int i=0;i<lump.nbr;i++) {
    //(bspf_vertex_t)
    m->ve[i]=lo[i];
    //m->ve[i].flag=0;
    //printf("%d %d %d\n",i,m->ve[i].faceop&((1<<24)-1),m->ve[i].indop);
  }

  free(lo);

  m->nve=lump.nbr;
  //printf("%d vertices\n",lump.nbr);

  return true;
}

static bool loadface(FILE *in, map_t *m,lump_t lump)
{
  if(m->fa) {
    PRINTLOG(1,"more that one face lump");
    free(m->fa);
  }
  m->fa=(face_t*)malloc(sizeof(face_t)*lump.nbr);
  if(m->fa==NULL) {
    PRINTLOG(1,"no enugth memory");
    return false;
  }

  bspf_face_t *lo=(bspf_face_t*)malloc(lump.nbr*sizeof(bspf_face_t));
  fread(lo,sizeof(bspf_face_t),lump.nbr,in);

  for(int i=0;i<lump.nbr;i++) {
    //(bspf_face_t)
    m->fa[i]=lo[i];
    //m->fa[i].flag=0;
    //printf("%d %d %d\n",i,m->fa[i].vertex,m->fa[i].nbrvertex);
  }

  free(lo);

  m->nfa=lump.nbr;


  return true;
}

static bool loadplane(FILE *in, map_t *m,lump_t lump)
{
  if(m->pl) {
    PRINTLOG(1,"more that one plane lump");
    free(m->pl);
  }
  m->pl=(plane_t*)malloc(sizeof(plane_t)*lump.nbr);
  if(m->pl==NULL) {
    PRINTLOG(1,"no enugth memory");
    return false;
  }

  bspf_plane_t *lo=(bspf_plane_t*)malloc(lump.nbr*sizeof(bspf_plane_t));
  fread(lo,sizeof(bspf_plane_t),lump.nbr,in);

  for(int i=0;i<lump.nbr;i++) {
    //(bspf_plane_t)
    m->pl[i]=lo[i];
    //m->pl[i].flag=0;
  }

  free(lo);

  m->npl=lump.nbr;

  return true;
}

static bool loadbspint(FILE *in, map_t *m,lump_t lump)
{
  if(m->bi) {
    PRINTLOG(1,"more that one bsp_int lump");
    free(m->bi);
  }
  m->bi=(bspint_t*)malloc(sizeof(bspint_t)*lump.nbr);
  if(m->bi==NULL) {
    PRINTLOG(1,"no enugth memory");
    return false;
  }
  fread(m->bi,sizeof(bspf_int_t),lump.nbr,in);
  return true;
}

static bool loadbspleaf(FILE *in, map_t *m,lump_t lump)
{
  if(m->bl) {
    PRINTLOG(1,"more that one bsp_leaf lump");
    free(m->bl);
  }
  m->bl=(bspleaf_t*)malloc(sizeof(bspleaf_t)*lump.nbr);
  m->nbl=lump.nbr;

  if(m->bl==NULL) {
    PRINTLOG(1,"no enugth memory");
    return false;
  }
  fread(m->bl,sizeof(bspf_leaf_t),lump.nbr,in);
  
  return true;
}

map_t *loadmap(char filename[])
{
  FILE *in;
  
  if((in=fopen(filename,"rb"))==NULL) {
    PRINTLOG(1,"cannot open file %s",filename);
    return NULL;
  }

  char id[5];
  int ver;
  short nbrbitnum;
  
  fread(id,1,4,in);
  fread((char*)&ver,4,1,in);
  fread((char*)&nbrbitnum,2,1,in);

  id[4]=0;
  if(strcmp(id,IDBSPF)) {
    PRINTLOG(1,"wrong ID in %s",filename);
    return NULL;
  }

  map_t *map=new map_t;

  lump_t lump;
  do {
    if(!loadlump(in,lump)) {
      PRINTLOG(1,"wrong LUMP ID in %s",filename);
      delete map;
      return NULL;
    }
    //printf("lump %d nbr %d size %d\n",lump.lump,lump.nbr,lump.size);
    switch(lump.lump) {
    case end: break;
    case vertex: loadvertex(in,map,lump); break;
    case face: loadface(in,map,lump); break;
    case plane: loadplane(in,map,lump); break;
    case bsp_leaf: loadbspleaf(in,map,lump); break;
    case bsp_int: loadbspint(in,map,lump); break;
    default: PRINTLOG(1,"unknown lump %d int %s",lump.lump,filename);
    }
  } while(lump.lump!=end);
  
  fclose(in);

  // FIX
  //check integrity
  if(map->ve == NULL || map->fa == NULL || map->pl==NULL || map->bi == NULL ||
     map->bl == NULL ) {
    PRINTLOG(1,"error integrity %s",filename);
    delete map;
    return NULL;
  }

  for(int i=0;i<map->nfa;i++) {
    for(int j=0;j<map->fa[i].nbrvertex;j++)
      map->ve[map->fa[i].vertex+j].face=i;
  }

  return map;
}
