#include "brush.hpp"
#include "vect.hpp"
#include "utils.hpp"
#include "plane.hpp"

#include <stdio.h>
#include <stdlib.h>
#include <iostream>

brush_t *brush_t::origBrush=NULL;
int brush_t::init=0;

// par defaut, brush = origBrush
brush_t::brush_t()
{
  if(init==0)
    initBrush();

  if(origBrush)
    *this=*origBrush;
}

brush_t::brush_t(const brush_t &b)
  :type(b.type),tab(b.tab)
{
}

brush_t::~brush_t()
{
}


void brush_t::print()
{
  unsigned int i;
  std::cout << tab.size() << " faces\n";
  
  for(i=0;i<tab.size();i++) {
    tab[i].print();
    tab[i].p.print();cout << " " << tab[i].aera() << endl;

  }
}

/***********/

// model brush
void brush_t::initBrush(void)
{
  init=1;
  origBrush=new brush_t;
  origBrush->tab.resize(6);
  
  vec4_t n;
  plane_t p;
  
  for(int i=0;i<6;i++) {
    n[0]=-(i==0)+(i==3);
    n[1]=-(i==1)+(i==4);
    n[2]=-(i==2)+(i==5);
    n[3]=minmaxb;
    vec4Copy(n,p.n);
    
    face_t *f=new face_t(&p);
    origBrush->tab[i]=*f;
    delete f;
  }
}

void brush_t::cutByPlane(plane_t *p)
{
  unsigned int i;
  face_t *f;
  
  /*check if p is a plane of b */

  // FIX operator ==
  /*
  for(i=0;i<tab.size();i++) {
    
    if(*p==tab[i].p) {
      // do nothing
      return;
    }
    else if(*p==(tab[i].p).opp()) {
      // this cut destroy the brush
      tab.resize(0);
      return;
    }
    } */

  /* create the new face */
  f=new face_t(p);
  
  for(i=0;i<tab.size();i++) {
    if(tab[i].checkIntersectFace(p)>-1) {
      tab[i].cutByPlane(p);
      f->cutByPlane(&(tab[i].p));
    }
  }
  
  //delete obsolete faces
  for(i=0;i<tab.size();i++)
    if(tab[i].tab.size()==0) {
      if(i<tab.size()-1)
	tab[i]=tab[tab.size()-1];
      tab.pop_back();
      i--;
    }
  
  if(f->tab.size()>0) {
    tab.push_back(*f);
  }
  
  delete f;
}

/* return 2 if p cut a 
   0 if a is on the positive side of p
   1 if a is on the negative side 
   or something like that */
int brush_t::checkIntersect(plane_t *p)
{
  unsigned int i,j;
  int k=0;
  int d;

  for(i=0;i<tab.size();i++)
    for(j=0;j<tab[i].tab.size();j++) {
      d=p->checkVertex(&(tab[i].tab[j]));
      if(d>0)
	k=k|1;
      if(d<0) 
	k=k|2;
      if(k==3) 
	return 2;
    }
  
  if(k<0 || k>3) Error(" ");
  return k-1;
}
