#include "bsp.hpp"
#include "portal.hpp"

bsp_t *searchbsp(bsp_t *b,vec3_t vv)
{
  if(b->leaf) {
    return b;
  } else {
    vertex_t v(vv);
    int d=b->plane.checkVertex(&v) ;
    if(d>=0) return searchbsp(b->child[0],vv);
    return searchbsp(b->child[1],vv);
  }
}

static void destroyBspPrt(bsp_t *b)
{
  if(b->leaf) {
    while(b->portals.size()>0) {
      portal_t *p=b->portals[0];
      p->remove();
      delete p;
    }
  } else {
    destroyBspPrt(b->child[0]);
    destroyBspPrt(b->child[1]);
  }
}

static void destroyBsp_rec(bsp_t *b)
{
  if(b->leaf==0) {
    destroyBsp_rec(b->child[0]);
    destroyBsp_rec(b->child[1]);
  }
  delete b;
}

void destroyBsp(hstr_t *h)
{
  destroyBspPrt(h->bsp);
  destroyBsp_rec(h->bsp);
  delete h->ext;
}

static unsigned int nbrprt=0;

static void saveBsp_rec(bsp_t *b,ofstream &ou)
{

  ou.write((char*)&(b->id),sizeof(b->id));
  ou.write((char*)&(b->num),sizeof(b->num));
  ou.write((char*)&(b->mark),sizeof(b->mark));
  ou.write((char*)&(b->leaf),sizeof(b->leaf));
  // type...

  b->plane.save(ou);

  if(b->leaf==0) {
    saveBsp_rec(b->child[0],ou);
    saveBsp_rec(b->child[1],ou);
  } else {
    unsigned int i=b->portals.size();
    ou.write((char*)&i,sizeof(i));
    for(i=0;i<b->portals.size();i++) {
      portal_t *p=b->portals[i];
      if(p->numsave==0) 
	p->numsave=++nbrprt;
      ou.write((char*)&p->numsave,sizeof(p->numsave));
      if(b==p->bsp[0]) {
	unsigned char c=0;
	ou.write((char*)&c,1);
	p->save(ou);
      } else {
	unsigned char c=1;
	ou.write((char*)&c,1);
      }
    }
  }
}

static vector<portal_t*> vport;

static portal_t *readport(unsigned int i)
{
  if(i>= vport.size()) return NULL;
  return vport[i];
}

static void setport(unsigned int i,portal_t *p)
{
  if(i>=vport.size()) {
    unsigned int j=vport.size();
    vport.resize(i+1);
    for(;j<vport.size();j++) 
      vport[j]=NULL;
  }
  vport[i]=p;
}

static bsp_t *loadBsp_rec(ifstream &f)
{
  bsp_t *b=new bsp_t;
  b->brush=NULL;
  f.read((char*)&(b->id),sizeof(b->id));
  f.read((char*)&(b->num),sizeof(b->num));
  f.read((char*)&(b->mark),sizeof(b->mark));
  f.read((char*)&(b->leaf),sizeof(b->leaf));
  // type...

  b->plane.load(f);

  if(b->leaf==0) {
    b->child[0]=loadBsp_rec(f);
    b->child[1]=loadBsp_rec(f);
  } else {
    unsigned int i,fin;
    f.read((char*)&fin,sizeof(fin));
    b->portals.resize(fin);
    for(i=0;i<fin;i++) {
      unsigned char c;
      unsigned int id;
#ifdef DEBUG
      //if(sizeof(id)!=sizeof(portal_t::numsave)) Error(" ");
#endif
      f.read((char*)&id,sizeof(id));
      //printf("id=%d\n",id);
      f.read((char*)&c,1);
      if(readport(id)==NULL)
	setport(id,new portal_t);
#ifdef DEBUG
      if(readport(id)==NULL)
	Error("lala");
#endif
      if(c==0) {
	readport(id)->load(f);
	readport(id)->bsp[0]=b;
      } else {
	readport(id)->bsp[1]=b;
      }
      b->portals[i]=readport(id);
    }
  }
  
  return b;
}

static void floodnumsave(bsp_t *b)
{
  if(b->leaf) {
    for(unsigned int i=0;i<b->portals.size();i++)
      b->portals[i]->numsave=0;
  } else {
    floodnumsave(b->child[0]);
    floodnumsave(b->child[1]);
  }
}

void saveBsp(hstr_t *h, char *f)
{
  floodnumsave(h->bsp);

  nbrprt=0;

  ofstream ou(f,fstream::binary|fstream::out);

  ou.write((char*)&(h->nbrleafin),sizeof(h->nbrleafin));
  ou.write((char*)&(h->nbrleaf),sizeof(h->nbrleaf));
  ou.write((char*)&(h->nbrnodeint),sizeof(h->nbrnodeint));
  
  saveBsp_rec(h->bsp,ou);
  saveBsp_rec(h->ext,ou);
  
  ou.close();
}

static void testbsp(bsp_t *b)
{
  if(b->leaf) {
    for(unsigned int i=0;i<b->portals.size();i++)
      if(b->portals[i]->bsp[0]==NULL)
	Error("portal->bsp[0]==NULL");
  } else {
    testbsp(b->child[0]);
    testbsp(b->child[1]);
  }
}


hstr_t *loadBsp(char *f)
{
  ifstream in(f,fstream::binary|fstream::in);
  
  hstr_t *h=new hstr_t;

  in.read((char*)&(h->nbrleafin),sizeof(h->nbrleafin));
  in.read((char*)&(h->nbrleaf),sizeof(h->nbrleaf));
  in.read((char*)&(h->nbrnodeint),sizeof(h->nbrnodeint));
  
  h->bsp=loadBsp_rec(in);
  h->ext=loadBsp_rec(in);

  vport.resize(0);
  
  in.close();

  testbsp(h->bsp);
  testbsp(h->ext);

  return h;
}
