#include "face.hpp"
#include <iostream>

#ifdef DEBUGMEM
unsigned long long int face_t::nbr_mem=0;
#endif

face_t::face_t()
  /*:of(NULL),op(NULL)*/
{
#ifdef DEBUGMEM
  nbr++;
  if(!(nbr%100000)) printf("nbr face_t : %Ld\n",nbr);
#endif
}

face_t::face_t(const face_t &a)
  :/*of(NULL),op(NULL),*/p(a.p),tab(a.tab),flags(a.flags),i0(a.i0),text(a.text),num(a.num)
{
#ifdef DEBUGMEM
  nbr++;
  if(!(nbr%100000)) printf("nbr face_t : %Ld\n",nbr);
#endif
}

face_t::~face_t()
{
#ifdef DEBUGMEM
  nbr--;
#endif
}

const face_t & face_t::operator =(const face_t &a)
{
  if(this != &a) {
    tab=a.tab;
    p=a.p;
    flags=a.flags;
    i0=a.i0;
    text=a.text;
    num=a.num;
  }
  //test();
  return *this;
}

void interEdgePlane_(vertex_t *a,vertex_t *b,vertex_t *c,plane_t *p)
{
  vertex_t u=*a,v=*b;
  vec4_t n;
  vec_t t;

  vec4Copy(p->n,n);
  
  vec3Sub_(u.v,v.v);
  
  vec_t de=vec3DotProduct(v.v,n);
  vec_t nu=-vec3DotProduct(u.v,n);

  if(de==0) Error("de==0");
  t=(nu-n[3])/de;
  
  vec3Scale_(t,v.v);
  vec3Add_(u.v,v.v);
  
  *c=v;

#if DEBUG
  if(p->checkVertex(c)) {
    p->print();
    cout << vec3DotProduct(a->v,p->n)+p->n[3] << endl;
    cout << vec3DotProduct(b->v,p->n)+p->n[3] << endl;
    cout << vec3DotProduct(c->v,p->n)+p->n[3] << endl;
    Error("checkvertex!=0 in interedgeplane");
  }
#endif
} 


face_t::face_t(plane_t *pl)
  :/*of(NULL),op(NULL),*/p(*pl),tab(4)
{
#ifdef DEBUGMEM
  nbr++;
  if(!(nbr%100000)) printf("nbr face_t : %Ld\n",nbr);
#endif

  unsigned int i,mi=0;
  vec_t m(0);
  int j;
  
  for(i=0;i<3;i++)
    if(i==0 || m<fabs(p.n[i])) {
      mi=i;
      m=fabs(p.n[i]);
    }
  
  i=mi;

  for(j=0;j<4;j++) {
    int j2=j;
    if(p.n[i]<0)
      j2=3-j;
    
    int x=0,y=0,z=0;
    switch(j2) {
    case 0: x=1;y=1;break;
    case 1: x=1;y=-1;break;
    case 2: x=-1;y=-1;break;
    case 3: x=-1;y=1;break;
    }
    
    switch(i) {
    case 0: z=x; x=-1;break;
    case 1: z=y; y=-1;break;
    case 2: z=x;x=y;y=z;z=-1;break;
    }
    
    vertex_t *v1= new vertex_t(x*minmaxb,y*minmaxb,z*minmaxb);

    switch(i) {
    case 0: x=1;break;
    case 1: y=1;break;
    case 2: z=1;break;
    }

    vertex_t *v2= new vertex_t(x*minmaxb,y*minmaxb,z*minmaxb);

    interEdgePlane_(v1,v2,&tab[j],pl);

    delete v1;
    delete v2;
  }

#ifdef DEBUG
  test();
#endif
}

void face_t::reverse()
{
  unsigned int i;
  vector_stat<hedge_t> t2(tab.size());
  for(i=0;i<tab.size();i++)
    t2[tab.size()-1-i]=tab[i];

  tab.exchange(t2);

  vec4Scale_(-1,p.n);
#ifdef DEBUG
  test();
#endif
}

void face_t::test() const
{
  unsigned int nbr=tab.size();
  vec3_t u,v;
  vec3_t n;
  unsigned int i=0;

  if(nbr<3) return;

  do {
    vec3Copy(tab[i+1].v,u);
    vec3Copy(tab[i+2].v,v);
    vec3Sub_(tab[i].v,u);
    vec3Sub_(tab[i].v,v);
    vec3CrossProduct(u,v,n);
    i++;
  } while(i+2<nbr && vec3LengthS(n)<EPSILON*EPSILON);

  if(i+2==nbr) Error("test()");

  //n[3]=-vec3DotProduct(n,[0]);
  
  if(vec3DotProduct(n,p.n)<0) {
    print();
    Error("test() normal");
  }

  for(i=0;i<nbr;i++) {
    if(p.checkVertex(&(tab[i]))) {
      p.print();printf("\n");
      tab[i].print();
      Error("test() check");
    }
  }
}

void face_t::print() const
{
  unsigned int j;
  
  std::cout << tab.size() << "\n";
  for(j=0;j<tab.size();j++) {
    tab[j].print();
  }
}


/* return 2 if p cut *this 
   0 if *this is on the positive side of p
   1 if *this is on the negative side 
   -1 if p==this->p
   or something like that */
int face_t::checkIntersectFace(plane_t *p) const
{
  unsigned int j;
  int d,k=0;
  unsigned int s=tab.size();

  for(j=0;j<s;j++) {
    d=p->checkVertex(&(tab[j]));
    if(d>0)
      k=k|1;
    if(d<0) 
      k=k|2;
    if(k==3) 
      return 2;
  }

  return k-1;
}

void face_t::cutByPlane(plane_t *pl)
{
  //static int debug=0;
  int nbr=tab.size();
  int i,j=0,k,ii,i2;
  int d=0,od;
  int flag=0;

  int da=0;

#ifdef DEBUG
  test();
#endif

  if(nbr==0) return;

  vector_stat<hedge_t> pp(nbr+1);
  
  for(i=0;(i<nbr) && (!flag);i++) {
    d=pl->checkVertex(&(tab[i]));
    flag|=(d>0);
  }

  if(flag==0) {
    tab.resize(0);
    return;
  }

  od=d;
  k=i+nbr;
  ii=i;

  for(;ii<k;ii++) {
    i=ii%nbr;
    i2=(ii+1)%nbr;
    d=pl->checkVertex(&(tab[i]));
    //if(debug)
    //  printf("d=%d\n",d);
    
    if(d>=0 && od>=0) {
      pp[j]=tab[i];
      j++;
    }

    if(d==0 && od<0) {
      pp[j]=tab[i];
      j++;
    }

    if(od<0 && d>0) {
      interEdgePlane_(&(tab[(i-1+nbr)%nbr]),&(tab[i]),&(pp[j]),pl);

#ifdef DEBUG
      if(da) Error("da>1");

      if(p.checkVertex(&(pp[j]))) {
	cout << vec3DotProduct(tab[i].v,p.n)+p.n[3] << endl;
	cout << vec3DotProduct(tab[(i-1+nbr)%nbr].v,p.n)+p.n[3] << endl;
	cout << vec3DotProduct(pp[j].v,p.n)+p.n[3] << endl;
	cout << vec3DotProduct(pp[j].v,pl->n)+pl->n[3] << endl;
	Error("face.cutbyplane checkvertex IN p");
      }

      if(pl->checkVertex(&(pp[j]))) {
	Error("face.cutbyplane checkvertex IN pl");
      }
#endif
      j++;

      pp[j]=tab[i];
      j++;

      da++;
    }
    
    if (d<0 && od>0) {
      interEdgePlane_(&(tab[(i-1+nbr)%nbr]),&(tab[i]),&(pp[j]),pl);
#ifdef DEBUG
      if(p.checkVertex(&(pp[j])))
	Error("face.cutbyplane checkvertex IN p");
      if(pl->checkVertex(&(pp[j]))) {
	Error("face.cutbyplane checkvertex IN pl");
      }
#endif
      j++;
    }

    od=d;
  }

#if 0
  //DEBUG
  for(i=0;i<j;i++) {
    if(p.checkVertex(&(pp[i]))) {
      printf("%f\n",vec3DotProduct(pp[i].v,p.n)+p.n[3]);
      p.print();printf("\n");
      pl->print();printf("\n");
      pp[i].print();
      //*(int*)NULL=0;
      Error("face.cutbyplane checkvertex POST");
    }
  }

  if(j<3) Error("j<3");
#endif

  pp.resize(j);
  
  tab.exchange(pp);

#ifdef DEBUG
  test();
#endif
}

void face_t::insert(unsigned int i,vertex_t &v)
{
  if(i==tab.size()-1) tab.push_back(v);
  else {
    // houlala
    tab.resize(tab.size()+1);
    for(unsigned int j=tab.size()-2;j>i;j--)
      tab[j+1]=tab[j];
    tab[i+1]=v;
  }
}


/**** SAVE LOAD ****/

void face_t::save(ofstream &f) const
{
  unsigned int i=tab.size();
  p.save(f);
  f.write((char*)&flags,sizeof(flags));
  f.write((char*)&i0,sizeof(i0));
  f.write((char*)&i,sizeof(i));
  f.write((char*)&num,sizeof(num));
  for(i=0;i<tab.size();i++)
    tab[i].save(f);
}
    
void face_t::load(ifstream &f) 
{
  unsigned int i;
  p.load(f);
  f.read((char*)&flags,sizeof(flags));
  f.read((char*)&i0,sizeof(i0));
  f.read((char*)&i,sizeof(i));
  f.read((char*)&num,sizeof(num));
  tab.resize(i);
  for(i=0;i<tab.size();i++)
    tab[i].load(f);

#ifdef DEBUG
  test();
#endif
}

void saveVectorFace(vector<face_t*> &v,char *f)
{
  ofstream ou(f,ofstream::binary|ofstream::out);
  unsigned int i=v.size();

  ou.write((char*)&i,sizeof(i));
  
  for(i=0;i<v.size();i++)
    v[i]->save(ou);

  ou.close();
}

void loadVectorFace(vector<face_t*> &v,char *f)
{
  ifstream in(f,ifstream::binary|ifstream::in);
  unsigned int i;

  in.read((char*)&i,sizeof(i));

  v.resize(i);
  for(i=0;i<v.size();i++) {
    v[i]=new face_t;
    v[i]->load(in);
  }

  in.close();
}

void destroyVectorFace(vector<face_t*> &v)
{
  while(v.size()) {
    face_t *f=v[v.size()-1];
    delete f;
    v.pop_back();
  }
}

double face_t::aera()
{
  if(tab.size()<3) return 0.;
  double r=0.;
  for(unsigned int i=1;i<tab.size()-1;i++) {
    vec3_t cp,a,b;
    vec3Copy(tab[i].v,a);
    vec3Sub_(tab[0].v,a);
    vec3Copy(tab[i+1].v,b);
    vec3Sub_(tab[0].v,b);
    vec3CrossProduct(a,b,cp);
    r+=vec3Length(cp)/2;
  }
  return r;
}



