#ifndef _integer_hpp_
#define _integer_hpp_

/* def classe int__t */
/* OK */

#include "build.hpp"

#ifdef GMP

#include <gmp.h>
#include <math.h>
#include <iosfwd>
#include <iostream>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>


#ifdef GMPMIX

#if 0
#define STATNBI
#define STATNBGI
#else
#define STATNBI    nbrop++
#define STATNBGI   nbrgop++
#endif

//#define MAXFF 1152921504606846976.
//#define MINGMP 0xF0000000
#define MAXFF 2305843009213693952.
#define MINGMP 0xE0000000
//#define MAXFF 4611686018427387904.
//#define MINGMP 0xC0000000

class int__t {
public:
  inline int__t();
  inline int__t(signed long long int ii);
  inline int__t(const int__t &);
  inline ~int__t();

  inline const int__t &operator =(const int__t &);

  inline bool operator ==(const int__t &) const;
  inline bool operator <(const int__t &) const;
  inline bool operator <=(const int__t &) const;
  inline int__t operator *(const int__t &) const;
  inline int__t operator +(const int__t &) const;
  inline int__t operator -() const;
  inline int__t operator /(const int__t &) const;

  inline bool operator !=(const int__t &) const;
  inline bool operator >(const int__t &) const;
  inline bool operator >=(const int__t &) const;
  inline int__t operator -(const int__t &) const;

  inline const int__t &operator +=(const int__t &);
  inline const int__t &operator -=(const int__t &);
  inline const int__t &operator *=(const int__t &);
  inline const int__t &operator /=(const int__t &);


  friend std::ostream & operator<<(std::ostream &o, const int__t &z);

  friend int__t gcd(const int__t &aa,const int__t &bb);
  friend int__t lcm(const int__t &aa,const int__t &bb);

  static void stat();

  inline double toDouble() const;
  

private:
  inline void reconvert() const;
  inline void convert() const;
  
  mutable float f;
  mutable signed long long int i;
  mutable mpz_t *mpz;
  mutable bool b; // 0 : mpz==NULL || i==mpz
  // 1 : mpz!=NULL && i!=mpz

  static unsigned long long int nbrop,nbrgop;
  static unsigned long long int nbr;
};


inline int__t::int__t()
  :f(0.),i(0),mpz(NULL),b(0)
{
#ifdef DEBUGMEM
  nbr++;
  if(!(nbr%1000000)) printf("nbr int__t : %Ld\n",nbr);
#endif
}

inline int__t::int__t(signed long long int ii)
  :f(ii),i(ii),mpz(NULL),b(0)
{
#ifdef DEBUGMEM
  nbr++;
  if(!(nbr%1000000)) printf("nbr int__t : %Ld\n",nbr);
#endif
}

inline int__t::int__t(const int__t &a)
  :f(a.f),i(a.i),b(a.b)
{
#ifdef DEBUGMEM
  nbr++;
  if(!(nbr%1000000)) printf("nbr int__t : %Ld\n",nbr);
#endif

  if(a.mpz==NULL) mpz=NULL;
  else {
    mpz=(mpz_t*)malloc(sizeof(mpz_t));
    mpz_init_set(*mpz,*(a.mpz));
  }
}  

inline const int__t &int__t::operator =(const int__t &a)
{
  STATNBI;
  if (this!=&a) {
    f=a.f;
    i=a.i;
    b=a.b;
    if(mpz) {
      mpz_clear(*mpz);
      free(mpz);
      mpz=NULL;
    }
    if(b && a.mpz) {
      STATNBGI;
      mpz=(mpz_t*)malloc(sizeof(mpz_t));
      mpz_init_set(*mpz,*(a.mpz));
    }
  }
  return *this;
}
 
inline int__t::~int__t()
{
#ifdef DEBUGMEM
  nbr--;
#endif
  if(mpz) {
    mpz_clear(*mpz);
    free(mpz);
  }
}

inline void int__t::convert() const
{
  mpz=(mpz_t*)malloc(sizeof(mpz_t));
  
  /*
  int64_t t=i;
  if(t<0) t=-t;
  if(t==0) 
  */
  
  mpz_init_set_si(*mpz,*(((int*)(void*)&i)+1));
  mpz_mul_ui(*mpz,*mpz,1<<16);
  mpz_mul_ui(*mpz,*mpz,1<<16);
  mpz_add_ui(*mpz,*mpz,*(((int*)(void*)&i)+0));
}

inline void int__t::reconvert() const
{
  //if(!mpz) Error("");
  //return;

  int s=(*mpz)->_mp_size;
  if(s<0) s=-s;
  if(s==0) {b=0;i=0;f=0;return;} //zero
  if(s>1) {
    for(int j=2;j<s;j++) {
      if((*mpz)->_mp_d[j]) return;  //cannot reconvert
    }
    
    if((*mpz)->_mp_d[1] & MINGMP) return;
  }

  b=0;
  i=(*mpz)->_mp_d[0];
  if(s>1)
    i+=((int64_t)(*mpz)->_mp_d[1]<<32);
  if((*mpz)->_mp_size<0) i=-i;
  f=i;
}

inline bool int__t::operator ==(const int__t &a) const
{
  STATNBI;
  if(b==0 && a.b==0)
    return i==a.i;
  STATNBGI;
  if(mpz==NULL) convert();
  if(a.mpz==NULL) a.convert();
  return (mpz_cmp(*mpz,*(a.mpz))==0);
}

inline bool int__t::operator <=(const int__t &a) const
{
  STATNBI;
  if(b==0 && a.b==0)
    return i<=a.i;
  STATNBGI;
  if(mpz==NULL) convert();
  if(a.mpz==NULL) a.convert();
  return (mpz_cmp(*mpz,*(a.mpz))<=0);
}

inline bool int__t::operator <(const int__t &a) const
{
  STATNBI;
  if(b==0 && a.b==0)
    return i<a.i;
  STATNBGI;
  if(mpz==NULL) convert();
  if(a.mpz==NULL) a.convert();
  return (mpz_cmp(*mpz,*(a.mpz))<0);
}

inline bool int__t::operator !=(const int__t &a) const
{
  return !(*this==a);
}

inline bool int__t::operator >(const int__t &a) const
{
  return !(*this<=a);
}

inline bool int__t::operator >=(const int__t &a) const
{
  return !(*this<a);
}

inline int__t int__t::operator *(const int__t &a) const
{
  STATNBI;
  if(b==0 && a.b==0) {
    float ff=f*a.f;
    if(fabs(ff)<MAXFF) {
      return int__t(i*a.i);
    }
  }
  STATNBGI;
  if(mpz==NULL) convert();
  if(a.mpz==NULL) a.convert();

  int__t r;
  
  r.b=1;
  r.f=f*a.f;
  r.mpz=(mpz_t*)malloc(sizeof(mpz_t));
  mpz_init(*(r.mpz));
  mpz_mul(*(r.mpz),*mpz,*(a.mpz));
  return r;
}

inline int__t int__t::operator +(const int__t &a) const
{
  STATNBI;
  if(b==0 && a.b==0) {
    float ff=f+a.f;
    if(fabs(ff)<MAXFF) {
      return int__t(i+a.i);
    }
  }
  STATNBGI;
  if(mpz==NULL) convert();
  if(a.mpz==NULL) a.convert();

  int__t r;
  
  r.b=1;
  r.f=f+a.f;
  r.mpz=(mpz_t*)malloc(sizeof(mpz_t));
  mpz_init(*(r.mpz));
  mpz_add(*(r.mpz),*mpz,*(a.mpz));
  r.reconvert();
  return r;
}

inline int__t int__t::operator -() const
{
  STATNBI;
  if(b==0) {
    return int__t(-i);
  }

  STATNBGI;

  int__t r;
  
  r.b=1;
  r.f=-f;
  r.mpz=(mpz_t*)malloc(sizeof(mpz_t));
  mpz_init(*(r.mpz));
  mpz_neg(*(r.mpz),*mpz);
  return r;
}

inline int__t int__t::operator -(const int__t &a) const
{
  return *this+(-a);
}

inline int__t int__t::operator /(const int__t &a) const
{
  STATNBI;
  if(b==0 && a.b==0) {
    //float ff=f/a.f;
    //if(fabs(ff)<MAXFF) {
    return int__t(i/a.i);
    //}
  }

  STATNBGI;
  if(mpz==NULL) convert();
  if(a.mpz==NULL) a.convert();

  int__t r;
  
  r.b=1;
  r.f=f/a.f;
  r.mpz=(mpz_t*)malloc(sizeof(mpz_t));
  mpz_init(*(r.mpz));
  mpz_cdiv_q(*(r.mpz),*mpz,*(a.mpz));

  r.reconvert();
  return r;
}

inline const int__t &int__t::operator +=(const int__t &a)
{
  return *this=*this+a;
}
inline const int__t &int__t::operator -=(const int__t &a)
{
  return *this=*this-a;
}
inline const int__t &int__t::operator *=(const int__t &a)
{
  return *this=*this*a;
}
inline const int__t &int__t::operator /=(const int__t &a)
{
  return *this=*this/a;
}

inline double int__t::toDouble() const
{
  if(b) return mpz_get_d(*mpz);
  else return static_cast<double>(i);
}


#else
 
typedef mpz_class int__t;

#endif


#else

typedef signed long long int__t;

#endif



typedef int__t ivec3_t[3];
typedef int__t ivec4_t[4];

extern int__t iabs(int__t);
extern int__t strToInt(char *);

#endif


