#include "map.hpp"
#include "utils.hpp"
#include "bsp.hpp"
#include "face.hpp"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <iostream>
#include <fstream>
using namespace std;

/*****************************/

static bsp_t *bspz=NULL;
static int zone=0;

// renvoie vrai si c est dans b
static bool isin(char *b,char c)
{
  int i;
  for(i=0;b[i];i++)
    if(c==b[i]) return true;
  return false;
}

// coupe la chaine de caractere (comme strtok_r...)
static char *strtok_r_(char *b,char *c, char **p)
{
  if(b)
    *p=b;

  if(*p==0) return NULL;

  while(**p && isin(c,**p)) (*p)++;
  char *p2=*p;
  while(**p && !isin(c,**p)) (*p)++;
  if(**p==0 && *p==p2) return NULL;
  if(**p!=0) {
    **p=0;
    (*p)++;
  }
  return p2;
}

void readMapBspZ(bsp_t *b,int z)
{
  bspz=b;
  zone=z;
}

static bool testbspz_r(bsp_t *b,const face_t *f,int z)
{
  if(b->leaf) {
    if(b->id!=z) return false;
    if(f->tab.size()>0) return true;
    return false;
  } else {
    if(f->checkIntersectFace(&b->plane)==-1) {
      return(testbspz_r(b->child[0],f,z) || testbspz_r(b->child[1],f,z));
    }
    
    face_t *fa=new face_t(*f);
    fa->cutByPlane(&b->plane);
    if(testbspz_r(b->child[0],fa,z)) {
      delete(fa);
      return true;
    }
    delete(fa);
    plane_t mp=b->plane.opp();
    fa=new face_t(*f);
    fa->cutByPlane(&mp);
    if(testbspz_r(b->child[1],fa,z)) {
      delete(fa);
      return true;
    }
    return false;
  }
}

static bool testbspzent_r(bsp_t *b,vertex_t vv,int z)
{
  if(b->leaf) {
    if(b->id!=z) return false;
    return true;
  } else {
    int d=b->plane.checkVertex(&vv) ;
    if(d>0) return testbspzent_r(b->child[0],vv,z);
    if(d<0) return testbspzent_r(b->child[1],vv,z);
    return testbspzent_r(b->child[0],vv,z) || testbspzent_r(b->child[1],vv,z);
  }
}

static bool testbspz(face_t *f)
{
  if(bspz==NULL) return true;
  return testbspz_r(bspz,f,zone);
}

static bool testbspzent(vec3_t v)
{
  if(bspz==NULL) return true;
  vertex_t vv;
  vec3Copy(v,vv.v);
  return testbspzent_r(bspz,vv,zone);
}

/********* READ MAP *************/

char *getStr(FILE *in)
{
  static char *pt=NULL; 
  static char tmp[1024];
  
  if(pt!=NULL) pt=strtok(NULL," \t\r\n");
  while(!feof(in) && ( pt==NULL || *pt==0 )) {
    fgets(tmp,1024,in);
    pt=strtok(tmp," \t\r\n");
  }
  if(feof(in)) pt=NULL;
  //if(pt) printf("\"%s\" \n",pt);
  //else printf("FIN\n");
  return pt;
}

static int cptbr;

void getBrush(map_t *m,FILE *in,vec_t tx,vec_t ty,vec_t tz)
{
  int i,j;//,ip;
  char *ptr=getStr(in);
  vec3_t pt[3];
  //int ok=1;

  //printf("getbrush\n");

  plane_t p;
  brush_t *b=new brush_t;
  
  b->type=TYPE_SOLID;
  
  //nbrfaces=0;
  while(strcmp(ptr,"}")) {
    for(i=0;i<3;i++) {
      if(strcmp(ptr,"(")) Error(" ");
      for(j=0;j<3;j++) {
	ptr=getStr(in);
	pt[i][j]=(int)atof(ptr);
      }
      pt[i][0]+=tx;
      pt[i][1]+=ty;
      pt[i][2]+=tz;

      ptr=getStr(in);
      if(strcmp(ptr,")")) Error(" ");
      ptr=getStr(in);
    }
    for(i=0;i<16;i++) ptr=getStr(in);
    
    pointsToPlane(pt,&p);
    //p.print();
    //ip=addPlane(&p);
    //printf("%d\n",ip);
    //if(ok)
    b->cutByPlane(&p);
    //if(ip==2) ok=0;
    //printBrush(b);
  }
  
  //add b to the map
  if(b->tab.size()>3) {
    //m->nbrbr++;
    //m->tabbr=(brush_t**)realloc(m->tabbr,(m->nbrbr)*sizeof(brush_t*));
    //m->tabbr[m->nbrbr-1]=b;
    m->tabbr.push_back(b);
  } else {
    Warn("Error in brush %d",cptbr);
  }
  cptbr++;
}

void getEntity(map_t *m,FILE *in,vec_t tx,vec_t ty,vec_t tz)
{
  char *ptr;
  
  while((ptr=getStr(in))) {
    if(!strcmp(ptr,"}")) return;
    if(!strcmp(ptr,"{")) getBrush(m,in,tx,ty,tz);
  }
  Error("Error in entity");
}

map_t *readMAP(map_t *m,const char *filename,int flag,vec_t tx,vec_t ty,vec_t tz)
{
  if(flag & READMAP_ZONE) 
    return m;
  
  FILE *in;
  char *pt;
  
  cptbr=0;
  
  //map_t *m;

  if(NULL==(in=fopen(filename,"rt"))) Warn("cannot open file %s",filename);
  //if(NULL==(in=fopen(filename,"rt"))) Error("cannot open file");

  //m=(map_t*)malloc(sizeof(map_t));

  //m->nbrbr=0;
  //  m->tabbr=(brush_t**)malloc(sizeof(brush_t*));
  
  while((pt=getStr(in))) {
    if(strcmp(pt,"{")) Error(" ");
    getEntity(m,in,tx,ty,tz);
  }

  printf("%s: %d brushes loaded\n",filename,cptbr);
  
  fclose(in);
  return m;
}

map_t *readTRUC(map_t *m,const char *filename,int flag,vec_t tx,vec_t ty,vec_t tz)
{
  FILE *in;
  char pt[1024];
  int i,n,j;
  char *p,*pok;
  vec3_t v;
  text_t text;

  int nbrfa=0;
  
  if(NULL==(in=fopen(filename,"rt"))) {
    Warn("cannot open file %s",filename);
    return m;
  }

  while(!feof(in)) {
    if(fgets(pt,1024,in)==NULL) continue;
    
    p=strtok_r(pt," ()\n\t",&pok);
    if(!p) continue;

    if(((!strcmp(p,"z")) && (flag & READMAP_ZONE)) 
       || ((!strcmp(p,"p") && (!(flag & READMAP_ZONE))))) {
      face_t *f=new face_t;
      // FIX
      if(flag & READMAP_ZONE)
	f->flags=FACE_ZONE;
      else {
	f->flags=FACE_BLOCK | FACE_DRAW;
	f->text=text;
      }

      p=strtok_r(NULL," ()\n\t",&pok);
      n=atoi(p);
      
      f->tab.resize(n);
      for(i=0;i<n;i++) {
	for(j=0;j<3;j++) {
	  p=strtok_r(NULL," ()\n\t",&pok);
	  v[j]=strToVec(p);
	}
	f->tab[i]=vertex_t(v[0]+tx,v[1]+ty,v[2]+tz);
      }
      
      // FIX trouver 3 vertex non colineaires
      vertexToPlane(&f->tab[0],&f->tab[1],&f->tab[2],&f->p);
      
      // test
      bool ok=true;
      for(i=0;i<n;i++)
	if(f->p.checkVertex(&(f->tab[i]))) {
	  Warn("readTruc: checkVertex in %s",filename);
	  ok=false;
	}
      if(ok) {
	if(!(flag & READMAP_ZONE)) {
	  if(testbspz(f)) { 
	    m->tabfa.push_back(f);
	    nbrfa++;
	  }
	}
	else { 
	  m->tabfa.push_back(f);
	  nbrfa++;
	  f->test();
	  //f->print();
	}
      }
    } else
    if((!strcmp(p,"e")) && (!(flag & READMAP_ZONE))) {
      p=strtok_r(NULL," ()\n\t",&pok);
      entit_t *e=new entit_t(p);
      m->tabent.push_back(e);
      for(j=0;j<3;j++) {
	p=strtok_r(NULL," ()\n\t",&pok);
	if(p==NULL) break;
	e->pos[j]=strToVec(p);
      }
      e->pos[0]+=tx;      e->pos[1]+=ty;      e->pos[2]+=tz;  

      
      if(p)
	for(j=0;j<3;j++) {
	  p=strtok_r(NULL," ()\n\t",&pok);
	  if(p==NULL) break;
	  e->dir[j]=strToVec(p);
	}

      if(!testbspzent(e->pos)) {
	delete e;
	m->tabent.pop_back();
      }
    } else 
    if(!strcmp(p,"t")) {
      char *p2=p;
      vec_t dx=0,dy=0,sx=1,sy=1;
      p=strtok_r(NULL," ()\n\t",&pok);
      dx=strToVec(p);
      p=strtok_r(NULL," ()\n\t",&pok);
      dy=strToVec(p);
      text=text_t(p2,dx,dy,sx,sy);
    }
  }
  
  if(nbrfa) printf("%s: %d faces loaded\n",filename,nbrfa);
  fclose (in);
  
  return m;
}


map_t *readSCR(map_t *m,const char *filename,int flag,vec_t tx,vec_t ty,vec_t tz)
{
  FILE *in;
  char pt[1024];
  
  if(NULL==(in=fopen(filename,"rt"))) {
    Warn("cannot open file %s",filename);
    return m;
  }

  while(!feof(in)) {
    if(fgets(pt,1024,in)==NULL) continue;
    
    if(pt[0]=='l') {
      // l "fichier.ext" tx ty tz (...)
      char f[100],*p;
      int i=0,j;
      vec_t Tx=0,Ty=0,Tz=0;

      while(pt[i] && pt[i] !='"') i++;
      if(!pt[i]) {
	Warn("erreur dans fichier scr");
	goto skipscr;
      }
      j=i+1;
      while(pt[j] && pt[j] !='"') j++;
      if(!pt[j]) {
	Warn("erreur dans fichier scr");
	goto skipscr;
      }
      strcpy(f,pt+i+1);
      f[j-i-1]=0;
      
      p=strtok(pt+j+1," \t\n");
      if(!p) goto goscr;
      Tx=strToVec(p);
      
      p=strtok(NULL," \t\n");
      if(!p) goto goscr;
      Ty=strToVec(p);

      p=strtok(NULL," \t\n");
      if(!p) goto goscr;
      Tz=strToVec(p);

    goscr:
      //cout << "scr: load " << f << " " << Tx << " " << Ty << " " << Tz << endl;
      readMap(m,f,flag,Tx+tx,Ty+ty,Tz+tz);
      
    skipscr:
      ;
    }
  }
  
  fclose (in);
  
  return m;
}

map_t *readMap(map_t *m,const char *filename,int flag,vec_t tx,vec_t ty,vec_t tz)
{
  if(m==NULL) {
    /*
      m=(map_t*)malloc(sizeof(map_t));
      m->tabbr=(brush_t**)malloc(sizeof(brush_t*));
      m->tabfa=(face_t**)malloc(sizeof(face_t*));
      m->nbrbr=m->nbrfa=0;
    */
    m=new map_t();
  }
  
  char *p3,ext[100];
  const char *p,*p2=filename;
  p=p2;
  while (*p) {
    if((*p)=='.')
      p2=p+1;
    p++;
  }
  strcpy(ext,p2);
  p3=ext;
  while(*p3) {
    *p3=tolower(*p3);
    p3++;
  }

  if(!strcmp(ext,"mscr")) {
    readSCR(m,filename,flag,tx,ty,tz);
    return m;
  }

  if(!strcmp(ext,"map")) {
    readMAP(m,filename,flag,tx,ty,tz);
    return m;
  }

  if(!strcmp(ext,"truc")) {
    readTRUC(m,filename,flag,tx,ty,tz);
    return m;
  }


  //if(!strcpy(ext,"ase")) {
  //  readASE(m,filename,tx,ty,tz);
  //}
    
  Warn("no known extention %s",ext);
    
  return m;
}


/*********** WRITE MAP *************/

#if 0
// probleme affichage de vec_t
void writeBrush(FILE *out,brush_t *b)
{
  int i,j;

  fprintf(out,"{\n");
  
  for(i=0;i<b->nbr;i++) {
    for(j=2;j>=0;j--) {
      fprintf(out,"( %.0f %.0f %.0f ) ",
	      b->tab[i].tab[j].v[0],
	      b->tab[i].tab[j].v[1],
	      b->tab[i].tab[j].v[2]);
    }
    fprintf(out,"AAATRIGGER [ 1 0 0 0 ] [ 0 -1 0 0 ] 0 1 1\n");
  }
  
  fprintf(out,"}\n");
}

void writeMap(map_t *m,const char *filename)
{
  int i;
  FILE *out;
  
  out=fopen(filename,"wt");

  fprintf(out,"{\n");
  fprintf(out,"%cclassname%c %cworldspawn%c\n",'"','"','"','"');
  fprintf(out,"%cmapversion%c %c220%c\n",'"','"','"','"');
  fprintf(out,"%cwad%c %c\\jeux\\half-life\\valve\\halflife.wad%c\n",'"','"','"','"');

  for(i=0;i<m->nbrbr;i++)
    writeBrush(out,m->tabbr[i]);

  fprintf(out,"}\n");

  fclose(out); 

}

#endif

void writeTruc(vector<face_t*> &t,char *filename,int lala)
{
  ofstream out(filename,ofstream::out); 

  if(!out.is_open()) Error("cannot open file");

  for(unsigned int i=0;i<t.size();i++) {
    face_t *f=t[i];

    if(lala) out << "p ";
    out << f->tab.size() << " ";
    if(!lala) 
      out << (double)(rand()%100)/100 << " ";
    for(unsigned int j=0;j<f->tab.size();j++) {
      out << f->tab[j].v[0] << " " << f->tab[j].v[1] << " " << f->tab[j].v[2] << " ";
    }
    out << endl;
  }

  out << "---\n";
  out.close();
}

/************ BRUSHES TO FACES *************/

void brushesToFaces(map_t *m)
{
  //cout << m->tabbr.size() << "\n";
  for(unsigned int i=0;i<m->tabbr.size();i++) {
    for(unsigned int j=0;j<m->tabbr[i]->tab.size();j++)
      m->tabfa.push_back(new face_t(m->tabbr[i]->tab[j]));
  }
  //cout << m->tabfa.size() << "\n";
}


