#include "plane.hpp"
#include "utils.hpp"
#include "build.hpp"

#include <stdio.h>
#include <stdlib.h>
#include <iostream>
//using namespace std;
#include "vect.hpp"

void pointsToPlane(const vec3_t pts[3],plane_t *p)
{
  vec3_t u,v,n;
  vec_t a;
   
  vec3Copy(pts[1],u);
  vec3Copy(pts[2],v);
  vec3Sub_(pts[0],u);
  vec3Sub_(pts[0],v);
  vec3CrossProduct(u,v,n);

  a=-vec3DotProduct(n,pts[0]);

  p->n[3]=a;
  vec3Copy(n,p->n);
}


void vertexToPlane(const vertex_t *v0, const vertex_t *v1, const vertex_t *v2, plane_t *p)
{
  vec3_t pt[3];
  
  vec3Copy(v0->v,pt[0]);
  vec3Copy(v1->v,pt[1]);
  vec3Copy(v2->v,pt[2]);
  
  pointsToPlane(pt,p);
}

#ifndef TESTEPSILON
// return 0 if v is on p, 1 if on the positive side, -1 if on the negative side
int plane_t::checkVertex(const vertex_t *v) const 
{
  vec_t d=vec3DotProduct(v->v,n)+n[3];
  
  if(d<0) return -1;
  if(d==0) return 0;
  return 1;
}

#else

int plane_t::checkVertex(const vertex_t *v) const 
{
  vec_t d=vec3DotProduct(v->v,n)+n[3];
  
  if(d<-EPSILON) return -1;
  if(d>EPSILON) return 1;
  return 0;
}

#endif

//FIX EPSILON ?
bool plane_t::operator==(const plane_t &f)
{
  //return vec4Equal(f.n,n);

  Warn("plane_t::operator==");

  int i;
  for(i=0;i<4;i++)
    if(n[i]!=0) {
      vec4_t t;
      vec_t k=f.n[i];
      k/=n[i];
      vec4Copy(n,t);
      vec4Scale_(k,t);
      return vec4Equal(f.n,t);
    }
  
  return true;
}

plane_t plane_t::opp() const 
{
  plane_t p;
  p.n[0]=-n[0];
  p.n[1]=-n[1];
  p.n[2]=-n[2];
  p.n[3]=-n[3];
  
  return p;
}

void plane_t::print() const
{
  std::cout << "(" << n[0] << " " << n[1] << " " << n[2] << " " << n[3] << ")";
}
