#include "pvs.hpp"
#include "vect.hpp"
#include "face.hpp"
#include <stdio.h>
#include <stdlib.h>
#include "pvsstab.hpp"

/*** monte carlo positive ****/

#define NBRTESTVERTEX 8

static void randomvertex(brush_t &b,vec3_t v)
{
  vec3_t m,M;
  unsigned int i,j,k;
  bool ok=false;

  for(i=0;i<b.tab.size();i++)
    for(j=0;j<b.tab[i].tab.size();j++) {
      for(k=0;k<3;k++) {
	if(!ok || m[k]>b.tab[i].tab[j].v[k])
	  m[k]=b.tab[i].tab[j].v[k];
	if(!ok || M[k]<b.tab[i].tab[j].v[k])
	  M[k]=b.tab[i].tab[j].v[k];
      }
      ok=true;
    }
  
  vertex_t vv;
  do {
    for(k=0;k<3;k++)
      vv.v[k]=m[k]+(M[k]-m[k])*((double)rand())/RAND_MAX;
    for(i=0;i<b.tab.size();i++)
      if(b.tab[i].p.checkVertex(&vv)<=0) break;
  } while(i!=b.tab.size());
  vec3Copy(vv.v,v);
}

bool traceseg(bsp_t *b,vec3_t aa,vec3_t bb)
{
  if(b->leaf) {
    if(b->id==1) return true;
    else return false;
  } else {
    vertex_t A,B,C;
    vec3Copy(aa,A.v);
    vec3Copy(bb,B.v);
    int d=b->plane.checkVertex(&A);
    int e=b->plane.checkVertex(&B);

    if(d>=0 && e>=0) return traceseg(b->child[0],aa,bb);
    if(d<=0 && e<=0) return traceseg(b->child[1],aa,bb);
    interEdgePlane_(&A,&B,&C,&b->plane);

    if(d>0)
      return traceseg(b->child[0],aa,C.v) && traceseg(b->child[1],bb,C.v);
    else
      return traceseg(b->child[1],aa,C.v) && traceseg(b->child[0],bb,C.v);
  }
}

/****************/

static char **tabvis;
static bsp_t **tabbsp;
static vec3_t (*tabtestvertex)[NBRTESTVERTEX];

static void visflood_leaf(bsp_t *b)
{
  vector<stabw_t*> vv;
  unsigned int i;

  tabvis[b->num&MASKNUMBSP][b->num&MASKNUMBSP]|=1;

  printf("visflood %d\n",b->num&MASKNUMBSP);
  for(i=0;i<b->portals.size();i++) 
    if(!(b->portals[i]->flag & PRT_BLOCK) &&
       !(b->portals[i]->flag & PRT_ZONE)) {
    bsp_t *bb;
    face_t *f=new face_t(b->portals[i]->face); 
    if(b==b->portals[i]->bsp[0])
      bb=b->portals[i]->bsp[1];
    else {
#ifdef DEBUG
      if(b!=b->portals[i]->bsp[1])
	Error("lelelele");
#endif
      bb=b->portals[i]->bsp[0];
      f->reverse();
    }
#ifdef DEBUG
    if(bb->id!=1) 
      Error("lili");
#endif
    stabw_t *ss=new stabw_t();
    ss->addface(f);
    ss->in=b;
    ss->out=bb;
    vv.push_back(ss);
    //printf("push %d\n",ss->f.size());
    tabvis[b->num&MASKNUMBSP][bb->num&MASKNUMBSP]|=1;
  }
  
  while(vv.size()>0) {
    stabw_t *s=vv[vv.size()-1];
    vv.pop_back();
    bsp_t *bb=s->out;
    for(i=0;i<bb->portals.size();i++) 
      if(!(bb->portals[i]->flag & PRT_BLOCK) &&
	 !(bb->portals[i]->flag & PRT_ZONE))
	if(&bb->portals[i]->face!=s->f[s->f.size()-1]) {
	  bsp_t *bbb;
	  face_t *f=new face_t(bb->portals[i]->face); 
	  if(bb==bb->portals[i]->bsp[0])
	    bbb=bb->portals[i]->bsp[1];
	  else {
#ifdef DEBUG
	    if(bb!=bb->portals[i]->bsp[1])
	      Error("lelelele");
#endif
	    bbb=bb->portals[i]->bsp[0];
	    f->reverse();
	  }
#ifdef DEBUG
	  if(bbb->id!=1) 
	    Error("lili");
#endif	  
	  stabw_t *ss=new stabw_t(*s);
	  ss->addface(f);
	  ss->out=bbb;
	  if(ss->vis()) {
	    vv.push_back(ss);
	    //printf("push %d %d\n",ss->f.size(),ss->out->num);
	    tabvis[b->num&MASKNUMBSP][bbb->num&MASKNUMBSP]|=1;
	  }
	  else {
	    delete ss;
	  }
	}
    delete s;
  }
}

static void createtabbsp(bsp_t *b)
{
  if(b->leaf) {
    if(b->id==1) 
      tabbsp[b->num&MASKNUMBSP]=b;
  } else {
    createtabbsp(b->child[0]);
    createtabbsp(b->child[1]);
  }
}

void visflood(hstr_t *h)
{
  int i,j,k;
  tabvis=(char**)malloc((1+h->nbrleafin)*sizeof(char*));
  tabbsp=(bsp_t**)malloc((1+h->nbrleafin)*sizeof(bsp_t*));
  tabtestvertex=(vec3_t (*)[NBRTESTVERTEX])malloc((1+h->nbrleafin)*sizeof(tabtestvertex[0]));
  for(i=1;i<=h->nbrleafin;i++) {
    tabbsp[i]=NULL;
    tabvis[i]=(char*)malloc(h->nbrleafin*sizeof(char));
    for(j=0;j<=h->nbrleafin;j++)
      tabvis[i][j]=0;
  }

  createtabbsp(h->bsp);

  for(i=1;i<=h->nbrleafin;i++) {
    for(j=0;j<NBRTESTVERTEX;j++)
      randomvertex(*tabbsp[i]->brush,tabtestvertex[i][j]);
  }

  for(i=1;i<=h->nbrleafin;i++) {
    tabvis[i][i]|=2;
    for(j=1;j<i;j++) {
      for(k=0;k<NBRTESTVERTEX*NBRTESTVERTEX;k++) {
	if(traceseg(h->bsp,tabtestvertex[i][k%NBRTESTVERTEX],tabtestvertex[j][k/NBRTESTVERTEX])) break;
      }
      if(k<NBRTESTVERTEX*NBRTESTVERTEX) {
	tabvis[i][j]|=2;
	tabvis[j][i]|=2;
      }
    }
  }
  
  for(i=1;i<=h->nbrleafin;i++)
    visflood_leaf(tabbsp[i]);
  
  for(i=1;i<=h->nbrleafin;i++) {
    for(j=1;j<=h->nbrleafin;j++)
      printf("%c",'0'+tabvis[i][j]);
    printf("\n");
  }
  
  for(i=1;i<=h->nbrleafin;i++)
    free(tabvis[i]);
  free(tabvis);
  free(tabbsp);
}
