#include "simplface.hpp"

class pbsp_t {
  public:
  pbsp_t():leaf(true),text(){}
  bool leaf;
  pbsp_t *child[2];
  face_t f;
  text_t text;
  plane_t p;
};

static void addpoly(pbsp_t &b, face_t &f)
{
  if(b.leaf) {
    unsigned int i,j;
    /*
    printf("addpoly leaf %f %f\n",f.aera(),b.f.aera());
    for(i=0;i<b.f.tab.size();i++)
      b.f.tab[i].print();
    printf("___\n");
    for(i=0;i<f.tab.size();i++)
      f.tab[i].print();
    printf("___\n");
    */

    for(i=0;i<f.tab.size();i++) {
      // si segment de f pas dans b.f -> split et rec
      for(j=0;j<b.f.tab.size();j++) {
	//si f[i] colineaire a b[j] : break
	if(( vertexOnLine(f.tab[i],
			  b.f.tab[j],b.f.tab[(j+1)%b.f.tab.size()]))
	   &&vertexOnLine(f.tab[(i+1)%f.tab.size()],
			  b.f.tab[j],b.f.tab[(j+1)%b.f.tab.size()])) break;
      }
      //si f[i] colinaire avec aucun cote de b 
      if(j==b.f.tab.size())  {
	//split
	//printf("split %d\n",i);
	//f.tab[i].print();
	//f.tab[(i+1)%f.tab.size()].print();
	vec3_t v;
	vec3Copy(f.tab[i].v,v);
	vec3Sub_(f.tab[(i+1)%f.tab.size()].v,v);
	vec3CrossProduct(b.f.p.n,v,b.p.n);
	b.p.n[3]=-vec3DotProduct(b.p.n,f.tab[i].v);
	b.p.norm();
	//printf("%f\n",vec3DotProduct(b.f.p.n,v));
#ifdef DEBUG
	if(b.p.checkVertex(&f.tab[i]) ||
	   b.p.checkVertex(&f.tab[(i+1)%f.tab.size()]))
	  Error("checkvertex");
#endif
	b.child[0]=new pbsp_t;	
	b.child[0]->f=b.f;
	b.child[0]->f.cutByPlane(&b.p);
	b.child[1]=new pbsp_t;
	b.child[1]->f=b.f;
	plane_t po=b.p.opp();
	b.child[1]->f.cutByPlane(&po);
	b.leaf=false;
	b.child[0]->text=b.text;
	b.child[1]->text=b.text;

	//appel rec
	addpoly(b,f);
	return;
      }
    }
    // ajout 
    //if(b.text.name !=NULL) Warn("superpose");
    if(b.text.name !=NULL && b.text!=f.text) {
      Warn("two differents textures");
      b.text.print(); printf(" "); f.text.print(); printf("\n");
    }
    b.text=f.text;
  } 
  // noeud interne
  else {
    //printf("addpoly %f %f\n",f.aera(),b.f.aera());
    face_t f1=f,f2=f;
    f1.cutByPlane(&b.p);
    plane_t po=b.p.opp();
    f2.cutByPlane(&po);
    
    if(f1.tab.size())
      addpoly(*b.child[0],f1);
    if(f2.tab.size())
      addpoly(*b.child[1],f2);
  }
}

static void destroy_rec(pbsp_t *b)
{
  if(!b->leaf) {
    destroy_rec(b->child[0]);
    destroy_rec(b->child[1]);
  }
  delete b;
}

/*** isnogood ***/

void recreate_rec(pbsp_t &b,portal_t &p)
{
  if(b.leaf) {
    text_t n;
    if(b.text!=n) {
      face_t *f=new face_t(b.f);
      f->text=b.text;
      p.listfaces.push_back(f);
    }
  } else {
    recreate_rec(*b.child[0],p);
    recreate_rec(*b.child[1],p);
  }
}

/*********/

static int nbrleaves(pbsp_t *b)
{
  if(!b->leaf)
    return nbrleaves(b->child[0])+nbrleaves(b->child[1]);
  else
    return 1;
}

static void mergeleaves(pbsp_t *b)
{
  if(!b->leaf) {
    mergeleaves(b->child[0]);
    mergeleaves(b->child[1]);
    if(b->child[0]->leaf && b->child[1]->leaf) 
      if(b->child[0]->text==b->child[1]->text) {
	b->text=b->child[0]->text;
	b->leaf=1;
	delete b->child[0];
	delete b->child[1];
      }
  }
}
static void simplify_portal(portal_t &p)
{
  if(p.listfaces.size()<2) return;

  //int lala=p.listfaces.size();

  unsigned int i;

  pbsp_t *b=new pbsp_t();
  b->f=p.face;
  b->f.flags=FACE_DRAW|FACE_BLOCK;
    
  // add et detruit faces portal
  for(i=0;i<p.listfaces.size();i++) {
    if(p.listfaces[i]->flags & FACE_DRAW) {
      addpoly(*b,*p.listfaces[i]);
      delete p.listfaces[i];
      p.listfaces[i]=p.listfaces[p.listfaces.size()-1];
      p.listfaces.pop_back();
      i--;
    }
  }

  //int n=p.listfaces.size();
  //int old=nbrleaves(b);

  // merge
  mergeleaves(b);

  // ajoute face protal
  recreate_rec(*b,p);
  //printf("simpl portal %d -> %d -> %d\n",lala-n,old,nbrleaves(b));

  // destroy bsp
  destroy_rec(b);
}

static void simplify_r(bsp_t *b)
{
  if(b->leaf) {
    for(unsigned int i=0;i<b->portals.size();i++) {
      portal_t *p=b->portals[i];
      if(p->bsp[0]==b)
	simplify_portal(*p);
    }
  } else {
    simplify_r(b->child[0]);
    simplify_r(b->child[1]);
  }
}

void simplify(hstr_t *h)
{
  printf("--- SIMPLIFY ---\n");

  simplify_r(h->bsp);

  // FIX inutile?
  for(unsigned int i=0;i<h->ext->portals.size();i++) {
    portal_t *p=h->ext->portals[i];
    if(p->bsp[0]==h->ext)
      simplify_portal(*p);
  }
}




