#ifndef _vect_hpp_
#define _vect_hpp_

#include <math.h>
#include <fstream>
using namespace std;
#include "build.hpp"

#ifdef VECRAT

#include "rat.hpp"
typedef rat_t vec_t;

#else

typedef double vec_t;

extern vec_t strToVec(char*);
extern void saveVec(ofstream &f,vec_t a);
extern vec_t loadVec(ifstream &f);
extern bool vecEqual(vec_t ,vec_t );

extern vec_t asqrt(vec_t a);

extern bool nanhugeval(vec_t a);

#endif

typedef vec_t vec4_t[4];
typedef vec_t vec3_t[3];
typedef vec_t vec2_t[2];

extern bool vec3EqualEps(const vec3_t u,const vec3_t v);


#define vec3Print(v) printf("(%f %f %f)",(v)[0],(v)[1],(v)[2])
#define vec4Print(v) printf("(%f %f %f %f)",(v)[0],(v)[1],(v)[2],(v)[3])

#define vec3Add_(v1,v2) {(v2)[0]+=(v1)[0];(v2)[1]+=(v1)[1];(v2)[2]+=(v1)[2];}
#define vec3Add(v1,v2,v3) {(v3)[0]=(v1)[0]+(v2)[0];(v3)[1]=(v1)[1]+(v2)[1];(v3)[2]=(v1)[2]+(v2)[2];}

#define vec3Sub_(v1,v2) {(v2)[0]-=(v1)[0];(v2)[1]-=(v1)[1];(v2)[2]-=(v1)[2];}
#define vec3Sub(v1,v2,v3) {(v3)[0]=(v1)[0]-(v2)[0];(v3)[1]=(v1)[1]-(v2)[1];(v3)[2]=(v1)[2]-(v2)[2];}

#define vec3CrossProduct(v1,v2,v3) { \
  (v3)[0]=(v1)[1]*(v2)[2]-(v1)[2]*(v2)[1]; \
  (v3)[1]=(v1)[2]*(v2)[0]-(v1)[0]*(v2)[2]; \
  (v3)[2]=(v1)[0]*(v2)[1]-(v1)[1]*(v2)[0];}\

#define vec3DotProduct(v1,v2) ((v1)[0]*(v2)[0]+(v1)[1]*(v2)[1]+(v1)[2]*(v2)[2])
#define vec4DotProduct(v1,v2) ((v1)[0]*(v2)[0]+(v1)[1]*(v2)[1]+(v1)[2]*(v2)[2]+(v1)[3]*(v2)[3])

#define vec3Copy(v1,v2) {(v2)[0]=(v1)[0];(v2)[1]=(v1)[1];(v2)[2]=(v1)[2];}
#define vec4Copy(v1,v2) {(v2)[0]=(v1)[0];(v2)[1]=(v1)[1];(v2)[2]=(v1)[2];(v2)[3]=(v1)[3];}

#define vec3Scale_(a,v1) {(v1)[0]*=(a);(v1)[1]*=(a);(v1)[2]*=(a);}
#define vec4Scale_(a,v1) {(v1)[0]*=(a);(v1)[1]*=(a);(v1)[2]*=(a);(v1)[3]*=(a);}

#define vec3Opp_(v) {(v)[0]=-(v)[0];(v)[1]=-(v)[1];(v)[2]=-(v)[2];}

#define vec3Null(v) {(v)[0]=(v)[1]=(v)[2]=0;}

#define vec3Length(v) (sqrt((v)[0]*(v)[0]+(v)[1]*(v)[1]+(v)[2]*(v)[2]))
#define vec3LengthS(v) ((v)[0]*(v)[0]+(v)[1]*(v)[1]+(v)[2]*(v)[2])

#define vec3Equal(u,v) ((u)[0]==(v)[0] && (u)[1]==(v)[1] && (u)[2]==(v)[2])
//#define vec3EqualEps(u,v) ((u)[0]==(v)[0] && (u)[1]==(v)[1] && (u)[2]==(v)[2])
#define vec4Equal(u,v) ((u)[0]==(v)[0] && (u)[1]==(v)[1] && (u)[2]==(v)[2] && (u)[3]==(v)[3])

#endif


