#include "build.hpp"
#include "vertex.hpp"
#include "vect.hpp"
#include <iostream>


vertex_t::vertex_t(vec_t x,vec_t y,vec_t z)
{
  v[0]=x;
  v[1]=y;
  v[2]=z;
}

void vertex_t::print() const
{
  //  std::cout << v[0] << " " << v[1] << " " << v[2] << " " << std::endl;
  printf("(%f %f %f)\n",v[0],v[1],v[2]);
}

bool vertex_t::operator==(vertex_t &o) const 
{
  return this->equal(&o);
}


// return u==v
#ifndef TESTEPSILON
bool vertex_t::equal(vertex_t *ve) const
{
  return vec3Equal(v,ve->v);
}
#else
bool vertex_t::equal(vertex_t *ve) const
{
  for(int i=0;i<3;i++) {
    vec_t d=v[i]-ve->v[i];
    if(d>EPSILON || d<-EPSILON)
      return false;
  }
  return true;
  //return vec3Equal(v,ve->v);
}

bool vertexOnSeg(vertex_t &v,vertex_t &a,vertex_t &b)
{
  vec3_t ba,va,p;
  vec3Copy(b.v,ba);
  vec3Sub_(a.v,ba);
  vec3Copy(v.v,va);
  vec3Sub_(a.v,va);
  vec_t t=vec3DotProduct(va,ba)/vec3DotProduct(ba,ba);
  if(t<0 || t>1) return false;
  vec3Copy(a.v,p);
  vec3Scale_(t,ba);
  vec3Add_(ba,p);
  vec3Sub_(v.v,p);
  t=vec3DotProduct(p,p);
  if(t<EPSILON*EPSILON) return true;
  //if(t<EPSILON) return true;
  return false;
}

bool vertexOnLine(vertex_t &v,vertex_t &a,vertex_t &b)
{
  vec3_t ba,va,p;
  vec3Copy(b.v,ba);
  vec3Sub_(a.v,ba);
  vec3Copy(v.v,va);
  vec3Sub_(a.v,va);
  vec_t t=vec3DotProduct(va,ba)/vec3DotProduct(ba,ba);
  //if(t<0 || t>1) return false;
  vec3Copy(a.v,p);
  vec3Scale_(t,ba);
  vec3Add_(ba,p);
  vec3Sub_(v.v,p);
  t=vec3DotProduct(p,p);
  if(t<EPSILON*EPSILON) return true;
  return false;
}

void saveVectorVertex(vector<vertex_t> &v,char *f)
{
  ofstream ou(f,ofstream::binary|ofstream::out);
  unsigned int i=v.size();

  ou.write((char*)&i,sizeof(i));
  
  for(i=0;i<v.size();i++)
    v[i].save(ou);

  ou.close();
}

void loadVectorVertexAppend(vector<vertex_t> &v,char *f)
{
  ifstream in(f,ifstream::binary|ifstream::in);
  unsigned int i;

  in.read((char*)&i,sizeof(i));
  
  unsigned int d=v.size();
  unsigned int s=i;
  v.resize(v.size()+i);
  for(i=0;i<s;i++) {
    v[d+i].load(in);
  }

  in.close();
}

#endif




