#include "readscript.hpp"
#include "log.hpp"
#include "cmd.hpp"
#include "pvar.hpp"
#include "function.hpp"
#include "timer.hpp"

#include <stdlib.h>
#include <string>

extern pvar_t programvars,globalvars;
extern functiontable_t functions;
typedef struct {
  void (*f)(int,char**);
  char *n;
} fcmd_t; 

static void echotest(int n,char **v)
{
  if(n) PRINTD("%s",v[0]);
}

static void letvar(int argc,char** argv)
{
  if(argc<2) {
    PRINTD("no many parameters");
    return;
  }
  
  if(globalvars.Search(argv[0])) {
    PRINTD("globalvar %s already exist",argv[0]);
    return;
  }
  
  pvar_type_t type=tint;

  if(!strcmp(argv[1],"double")) type=tdouble;

  globalvars.Add(argv[0],type,0);
}

static void setvar(int argc,char** argv)
{
  if(argc<2) {
    PRINTD("no many parameters");
    return;
  }

  if(programvars.Search(argv[0])) {
    programvars.Set(argv[0],argv[1]);
    return;
  }
  
  if(globalvars.Search(argv[0])) {
    globalvars.Set(argv[0],argv[1]);
    return;
  }

  PRINTD("var %s not found",argv[0]);
}

static void getvar(int argc,char** argv)
{
  char *v;
  
  if(argc<1) {
    PRINTD("no many parameters");
    return;
  }

  if(programvars.Search(argv[0])) {
    v=programvars.Get(argv[0]);
    if(v)
      PRINTD("programvar %s = %s",argv[0],v);
    free(v);
    return;
  }
  
  if(globalvars.Search(argv[0])) {
    v=globalvars.Get(argv[0]);
    if(v)
      PRINTD("globalvar %s = %s",argv[0],v);
    free(v);
    return;
  }

  PRINTD("var %s not found",argv[0]);
}

static void loadscript(int argc,char** argv)
{
  if(argc<1) {
    PRINTD("no many parameters");
    return;
  }

  readscriptfile(argv[0]);
}

static void printfct(int argc,char** argv)
{
  PRINTD("not implemented");
}

static void execfct(int argc,char** argv)
{
  if(argc<1) {
    PRINTD("no many parameters");
    return;
  }

  execfunction(argv[0],NULL);
}

static void printlist(int argc, char ** argv)
{
  if(argc>0) {
    if(argv[0][0]=='p')
      programvars.PrintList();
    if(argv[0][0]=='g')
      globalvars.PrintList();
    if(argv[0][0]=='f')
      functions.PrintList();
  } else {
    programvars.PrintList();
    globalvars.PrintList();
  }
}

static void printtimer(int argc, char **argv)
{
  timer_elm_t *p=globaltimer.tl.next;
  while(p) {
    PRINTD("(%d) %d: %s",p->id,p->time,p->fct);
    p=p->next;
  }
}

extern int nfcmd;
extern fcmd_t fcmd[];

static void printcmd(int argc,char **argv)
{
  for(int i=0;i<nfcmd;i++)
    PRINTD("%s",fcmd[i].n);
}

fcmd_t fcmd[]={{setvar,"set"},{getvar,"get"},{echotest,"echo"},{letvar,"let"},{loadscript,"script"},{printfct,"printfct"},{printlist,"list"},{execfct,"exec"},{printtimer,"timers"},{printcmd,"help"}};
int nfcmd=sizeof(fcmd)/sizeof(fcmd[0]);

/// delete fct , var
//// modifie fct

void executecmd(char* cmd)
{
  // FIXME   chaine entre " 

  int argc=0,i;
  char* argv[128];
  char* ptr;
  
  ptr=strtok(cmd," \n");

  while(ptr && argc<128) {
    argv[argc]=ptr;
    argc++;
    ptr=strtok(0," \n");
  }
  
  if(argc==0) return;

  for(i=0;i<nfcmd;i++)
    if(!strcmp(argv[0],fcmd[i].n)) {
      fcmd[i].f(argc-1,argv+1);
      return;
    }

  //PRINTD("command %s not found",argv[0]);

  getvar(1,argv);
}


  
  
