#include "function.hpp"
#include "scr_val.hpp"
#include "scr_var.hpp"
#include "log.hpp"
#include "fctbuildin.hpp"

#include <string>

void functiontable_t::hfree(void *p)
{
  //FIXME
  //Delete
}


functiontable_t::functiontable_t()
  :ht(functiontable_t::hfree,771)
{
}

functiontable_t::~functiontable_t()
{
}

void functiontable_t::Add(char* name, scr_fct_t* fct)
{
  //scr_fct_t *t=new scr_fct_t;
  //t->code=strdup(code);

  scr_fct_t* p=(scr_fct_t*)ht.Search(name);
  if(p) {
    PRINTD("function add: redefine function %s",name);
    //Delete(name);
  }
  
  ht.Add(name,(void*)fct);
}

void functiontable_t::PrintList(void)
{
  ht.PrintList();
}

scr_fct_t *functiontable_t::Get(char *n)
{
  return (scr_fct_t*)ht.Search(n);
}


/* execution */

static scr_val_t scr_execute_instrs(pvar_t *,scr_list_t<scr_instr_t*> *instrs);
static scr_val_t scr_execute_instr(pvar_t *,scr_instr_t *instr);
static scr_val_t scr_execute_expr(pvar_t *,scr_expr_t *expr);

static scr_val_t scr_execute_instrs(pvar_t *lv,scr_list_t<scr_instr_t*> *instrs)
{
  scr_val_t r;
  scr_list_t<scr_instr_t*> *p=instrs;
  while(p) {
    r=scr_execute_instr(lv,p->info);
    /* check return */
    if(r.type!=NOTYPE) // return
      return r;
    p=p->next;
  }
  return r; // notype
}


static scr_val_t scr_execute_instr(pvar_t *lv,scr_instr_t *instr)
{
  scr_val_t r;

  switch(instr->type) {
  case INBLOC: 
    r=scr_execute_instrs(lv,instr->bloc); 
    return r;
    break; 
  case INEXPR:
    scr_execute_expr(lv,instr->expr);
    return r;
    break;
  case INIF: 
    if(scr_execute_expr(lv,instr->expr)) r=scr_execute_instr(lv,instr->ins0); else 
    if(instr->ins1) r=scr_execute_instr(lv,instr->ins1);
    return r;
    break; 
  case INWHILE: 
    while(scr_execute_expr(lv,instr->expr)) {
      r=scr_execute_instr(lv,instr->ins0); 
      if(r.type!=NOTYPE)
	return r;
    }
    return r;
    break;
  case INDOWHILE: 
    do {
      r=scr_execute_instr(lv,instr->ins0); 
      if(r.type!=NOTYPE)
	return r;
    } while(scr_execute_expr(lv,instr->expr)); 
    return r;
    break;
  case INDECL:
    scr_registervar(lv,instr->name,instr->typedecl);
    if(instr->expr)
      scr_setvar(lv,instr->name,scr_execute_expr(lv,instr->expr));
    return r;
    break;
  case INRET:
    if(instr->expr)
      r=scr_execute_expr(lv,instr->expr);
    else { //return;
      r.type=TYPEINT;
      r.v.i=0;
    }
    return r;
    break;
  default: 
    PRINTD("error in function scr_execute_instr: default case");
    return r;
  }
}

static scr_val_t scr_execute_expr(pvar_t *lv,scr_expr_t *expr)
{
  //PRINTD("exectue e");
  //scr_val_t::printoper(expr->op);printf("\n");
  scr_val_t r;
  switch(expr->op) {
  case OPVAR:
    //PRINTD(expr->name);
    r=scr_getvar(lv,expr->name);
    if(r.type==NOTYPE) { //not found
      PRINTD("var not found");
      r.name=NULL;
      r.type=TYPEINT;
      r.v.i=0;
    } else {
      r.name=strdup(expr->name);
    }
    return r; //scr_val_t(expr); 
    break;
  case OPCONF:
  case OPCONI: 
  case OPCONS: 
    return scr_val_t(expr); 
    break;
  case OPFCT:
    if(expr->args) {
      scr_list_t<scr_val_t> *a,*la;
      scr_list_t<scr_expr_t*> *b=expr->args;
      while(b) {
	if(b==expr->args) {
	  la=a=new scr_list_t<scr_val_t>;
	}
	else {
	  a->next=new scr_list_t<scr_val_t>;
	  a=a->next;
	}
	a->info=scr_execute_expr(lv,b->info);
	b=b->next;
      }
      a->next=NULL;
      
      r=execfunction(expr->name,la);
      
      while(la) {
	a=la->next;
	delete la;
	la=a;
      }
    }
    return r;
    break; 
  case OPCAST:
    r=scr_execute_expr(lv,expr->u.e.opd0);
    r.transtype(expr->type);
    return r;
    break;
  default: // operator
    switch(scr_val_t::arity(expr->op)) {
    case 1:
      return scr_val_t::oper(lv,expr->op,scr_execute_expr(lv,expr->u.e.opd0));
    case 2:
      return scr_val_t::oper(lv,expr->op,scr_execute_expr(lv,expr->u.e.opd0),scr_execute_expr(lv,expr->u.e.opd1));
    default:
      PRINTD("error in function scr_execute_expr operator: default case");
      return scr_val_t(0);
    }
    break;
  }
}

scr_val_t execfunction(char *name, scr_list_t<scr_val_t>*a)
{
  int i=0; 

  while(i<nbrfctbuildin) {
    if(strcmp(name,fctbuildin[i].n)==0) {
      return (*fctbuildin[i].f)(a);
      break;
    }
    i++;
  }
  
  scr_fct_t *f=functions.Get(name);
  if(f==NULL) {
    PRINTD("function not found");
    return scr_val_t();
  }
  
  pvar_t localvars(7); // petite table pour ne pas perdre de temps a la destruction
  scr_list_t<scr_instr_t*> *arg=f->args;
  scr_val_t r;
  scr_list_t<scr_val_t>*la=a;

  // args
  while(la && arg) {
    scr_registervar(&localvars,arg->info->name,arg->info->typedecl);
    scr_setvar(&localvars,arg->info->name,la->info);
    arg=arg->next;
    la=la->next;
  }
  
  while(arg) {
    scr_registervar(&localvars,arg->info->name,arg->info->typedecl);
    scr_setvar(&localvars,arg->info->name,r);
    arg=arg->next;
  }
  
  r=scr_execute_instr(&localvars,f->instr);

  // FIXME cast

  return r;
}

