#ifndef _function_hpp_
#define _function_hpp_

#include "hashtable.hpp"
#include <stdlib.h>
#include "scr_val.hpp"

/* todo : enum ?*/
//special
#define NOOP 0
#define OPVAR 1
#define OPFCT 2
#define OPCAST 3
#define OPCONI 5
#define OPCONF 6
#define OPCONS 7


//affect
#define OPAFF 10 // =

//unary
#define OPUM  50 // - 
#define OPUN  51 // !

//binary
#define OPMUL 60
#define OPDIV 61
#define OPMOD 62
#define OPADD 63
#define OPSUB 64
#define OPL   65
#define OPG   66
#define OPLE  67
#define OPGE  68
#define OPEQ  69
#define OPNEQ 70
#define OPAND 71
#define OPOR  72

/* instructions */
#define NOINSTR 0
#define INBLOC  1

#define INEXPR  10
#define INDECL  11

#define INIF    20
#define INWHILE  21
#define INDOWHILE  22

#define INRET 30

/* type */
#define NOTYPE 0
#define TYPEVOID 1
#define TYPEINT 2
#define TYPEDOUBLE 3
#define TYPESTRING 4

/* special */
#define TYPEBREAK 128

//mask
#define TYPEGLOBAL 256


template <class T> struct scr_list_t {
  scr_list_t<T> *next;
  T info;
};

struct scr_expr_t {
  scr_expr_t() {name=NULL;args=NULL;u.e.opd0=u.e.opd1=NULL;}
  
  int op;
  char *name;
  int type;
  scr_list_t<scr_expr_t*> *args;
  union {
  // expr
    struct {
      scr_expr_t *opd0;
      scr_expr_t *opd1;
    } e;
    // consts
    int i;
    double d;
    char *s;
  } u;
};

struct scr_instr_t {
  scr_instr_t() {name=NULL;expr=NULL;ins0=ins1=NULL;bloc=NULL;}
  int type;     // type instruction
  int typedecl; // type for declaration
  char *name;   // declaration
  scr_expr_t *expr;  // expression & condition & return
  scr_instr_t *ins0; // { ... } & if true & while & do while
  scr_instr_t *ins1; // if false
  scr_list_t<scr_instr_t*> *bloc; //bloc
};

struct scr_fct_t {
  scr_fct_t(){name=NULL;args=NULL;instr=NULL;}
  int type;
  char *name;
  scr_list_t<scr_instr_t*> *args;
  int nargs;
  scr_instr_t *instr;
};


/* HASHTABLE */


class functiontable_t {
public:
  functiontable_t();
  ~functiontable_t();
  
  void Add(char* name,scr_fct_t *fct);
  //Delete(char* name);
  
  void PrintList(void);

  scr_fct_t *Get(char *);
 
protected:
  static void hfree(void *);

  hashtable_t ht;
};

// FIXME
extern functiontable_t functions;

scr_val_t  execfunction(char *name, scr_list_t<scr_val_t>*a);

#endif
