//#include "hashtable.hpp"
#include "pvar.hpp"
#include "log.hpp"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

void pvar_t::hfree(void *p)
{
  pvar_elm_t *t=(pvar_elm_t*)p;

  if(((t->flags)|PVAR_PROG)==0)  
    free(t->ptr);
  delete t;
}


pvar_t::pvar_t(int Size):ht(pvar_t::hfree,Size)
{
}  

pvar_t::~pvar_t()
{
}

void pvar_t::Add(char *name, pvar_type_t type, int flags,void *ptr)
{
  pvar_elm_t *p=new pvar_elm_t;

  p->flags=0;

  if(ptr!=NULL && type==tstring) 
    p->flags=PVAR_RO;
  
  if(ptr==NULL || type==tstring) {
    char *p=(char*)ptr;
    ptr=malloc(8);
    switch(type) {
    case tstring:
      ptr=(void*)p;
      if(p==NULL)
	ptr=(void*)strdup("");
      break;
    case tint:
      *(int*)ptr=0;
      break;
    case tdouble:
      *(double*)ptr=0.;
      break;
    default:
      break;
    }
  }
  else 
    p->flags|=PVAR_PROG;
  
  p->type=type;
  p->ptr=ptr;
  p->flags|=flags;

  ht.Add(name,(void*)p);
}

void pvar_t::Set(char *name, char *value)
{
  pvar_elm_t *p=(pvar_elm_t*)ht.Search(name);
  
  if(p==NULL) {
    PRINTDT(1,"pvar '%s' not found\n",name);
  } else if(p->flags & PVAR_RO) {
    PRINTDT(1,"pvar '%s' is read only\n",name);
  } else {
    switch(p->type) {
    case tchar:
      *(unsigned char*)(p->ptr)=atoi(value); break;
    case tint:
      *(int*)(p->ptr)=atoi(value); break;
    case tfloat:
      *(float*)(p->ptr)=atof(value); break;
    case tdouble:
      *(double*)(p->ptr)=atof(value); break;
    case tstring:
      if(p->ptr) free(p->ptr);
      (p->ptr)=(void*)strdup(value);break;
    default:
      break;
    }
  }
}

void pvar_t::Seti(char *name, int value)
{
  pvar_elm_t *p=(pvar_elm_t*)ht.Search(name);
  
  if(p==NULL) {
    PRINTDT(1,"pvar '%s' not found\n",name);
  } else if(p->flags & PVAR_RO) {
    PRINTDT(1,"pvar '%s' is read only\n",name);
  } else {
    switch(p->type) {
    case tchar:
      *(unsigned char*)(p->ptr)=value; break;
    case tint:
      *(int*)(p->ptr)=value; break;
    case tfloat:
      *(float*)(p->ptr)=value; break;
    case tdouble:
      *(double*)(p->ptr)=value; break;
    case tstring:
      if(p->ptr) free(p->ptr);
      char *r;
      asprintf(&r,"%d",value);
      p->ptr=(void*)r;
      break;
    default:
      break;
    }
  }
}

void pvar_t::Setf(char *name, double value)
{
  pvar_elm_t *p=(pvar_elm_t*)ht.Search(name);
  
  if(p==NULL) {
    PRINTDT(1,"pvar '%s' not found\n",name);
  } else if(p->flags & PVAR_RO) {
    PRINTDT(1,"pvar '%s' is read only\n",name);
  } else {
    switch(p->type) {
    case tchar:
      *(unsigned char*)(p->ptr)=(int)value; break;
    case tint:
      *(int*)(p->ptr)=(int)value; break;
    case tfloat:
      *(float*)(p->ptr)=value; break;
    case tdouble:
      *(double*)(p->ptr)=value; break;
    case tstring:
      if(p->ptr) free(p->ptr);
      char *r;
      asprintf(&r,"%f",value);
      p->ptr=(void*)r;
      break;
    default:
      break;
    }
  }
}

int pvar_t::Search(char *name)
{
  if(ht.Search(name)==NULL)
    return 0;
  else
    return 1;
}

pvar_type_t pvar_t::GetType(char *name)
{
  pvar_elm_t *p;
  if((p=(pvar_elm_t*)ht.Search(name))!=NULL) {
    return p->type;
  }
  return tnone;
}

char *pvar_t::Get(char *name)
{
  char *s;

  pvar_elm_t *p=(pvar_elm_t*)ht.Search(name);

  if(p==NULL) {
    PRINTDT(1,"pvar '%s' not found\n",name);
    return NULL;
  } 

  //s=(char*)malloc(20);
  switch(p->type) {
  case tchar:
    asprintf(&s,"%d",*(unsigned char*)(p->ptr)); break;
  case tint:
    asprintf(&s,"%d",*(int*)(p->ptr)); break;
  case tfloat:
    asprintf(&s,"%f",*(float*)(p->ptr)); break;
  case tdouble:
    asprintf(&s,"%f",*(double*)(p->ptr)); break;
  case tstring:
    asprintf(&s,"%s",(char*)(p->ptr)); break;
  default:
    break;
  }
  return s;
}

int pvar_t::Geti(char *name)
{
  pvar_elm_t *p=(pvar_elm_t*)ht.Search(name);
  
  if(p==NULL) {
    PRINTDT(1,"pvar '%s' not found\n",name);
    return 0;
  } 
  
  switch(p->type) {
  case tchar:
    return *(unsigned char*)(p->ptr);
  case tint:
    return *(int*)(p->ptr);
  case tfloat:
    return (int)*(float*)(p->ptr);
  case tdouble:
    return (int)*(double*)(p->ptr);
  case tstring:
    return atoi((char*)(p->ptr));
  default:
    break;
  }
  return 0;
}

double pvar_t::Getf(char *name)
{
  pvar_elm_t *p=(pvar_elm_t*)ht.Search(name);
  
  if(p==NULL) {
    PRINTDT(1,"pvar '%s' not found\n",name);
    return 0;
  } 
  
  switch(p->type) {
  case tchar:
    return (double)*(unsigned char*)(p->ptr);
  case tint:
    return (double)*(int*)(p->ptr);
  case tfloat:
    return *(float*)(p->ptr);
  case tdouble:
    return *(double*)(p->ptr);
  case tstring:
    return atof((char*)(p->ptr));
    default:
      break;
  }
  return 0;
}

void pvar_t::PrintList(void)
{
  ht.PrintList();
}
