#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "scr_val.hpp"
#include "scr_var.hpp"
#include "log.hpp"
#include "function.hpp"

scr_val_t::scr_val_t(int i)
{
  name=NULL;
  type=TYPEINT;
  v.i=i;
}

scr_val_t::scr_val_t()
{
  name=NULL;
  type=NOTYPE;
}

scr_val_t::scr_val_t(scr_expr_t *e)
{
  name=NULL;
  type=TYPEINT;
  v.i=0;
  switch(e->op) {
    /*case OPVAR:
    (*this)=scr_getvar(e->name);
    if(type==NOTYPE) { //not found
      PRINTD("not found");
      name=NULL;
      type=TYPEINT;
      v.i=0;
    } else {
      name=strdup(e->name);
    }
    break;
    */
  case OPCONI:
    name=NULL;
    type=TYPEINT;
    v.i=e->u.i;
    break;
  case OPCONF:
    name=NULL;
    type=TYPEDOUBLE;
    v.d=e->u.d;
    break;
  case OPCONS:
    name=NULL;
    type=TYPESTRING;
    v.s=strdup(e->u.s);
    break;
  default:
    PRINTD("error in scr_val_t contructor from expression: default case");
    name=NULL;
    type=TYPEINT;
    v.i=0;
    break;
  }
}

scr_val_t::scr_val_t(const scr_val_t &c)
{
  if(c.name!=NULL) name=strdup(c.name);
  else name=NULL;
  type=c.type;
  switch(type) {
  case TYPEINT: v.i=c.v.i; break;
  case TYPEDOUBLE: v.d=c.v.d; break;
  case TYPESTRING: v.s=strdup(c.v.s); break;
  }
}

scr_val_t::~scr_val_t()
{
  if(type==TYPESTRING) free(v.s);
  if(name) free(name);
}

const scr_val_t &scr_val_t::operator =(const scr_val_t &c)
{
  if(this!=&c) {
    if(type==TYPESTRING) free(v.s);
    if(name) free(name);
    if(c.name) name=strdup(c.name);
    type=c.type;
    switch(type) {
    case TYPEINT: v.i=c.v.i; break;
    case TYPEDOUBLE: v.d=c.v.d; break;
    case TYPESTRING: v.s=strdup(c.v.s); break;
    }
  }
  return *this;
}

scr_val_t::operator bool()
{
  if(type==TYPEINT) {
    return v.i;
  }
  if(type==TYPEDOUBLE) {
    return 1;
  }
  if(type==TYPESTRING) {
    return strlen(v.s)>0;
  }
  return 0;
}

void scr_val_t::transtype(int t)
{
  if(type!=t) {
    if(type==TYPEINT) {
      if(t==TYPEDOUBLE) 
	v.d=(double)v.i;
      if(t==TYPESTRING) 
	asprintf(&(v.s),"%d",v.i);
    }
    if(type==TYPEDOUBLE) {
      if(t==TYPEINT) 
	v.i=(int)v.d;
      if(t==TYPESTRING) 
	asprintf(&(v.s),"%f",v.d);
    }
    if(type==TYPESTRING) {
      char *s=v.s;
      if(t==TYPEINT)
	v.i=atoi(s);
      if(t==TYPEDOUBLE)
	v.d=atof(s);
      free(s);
    }
  }
  type=t;
}

int scr_val_t::arity(int op)
{
  if(op>=10 && op<50) return 2; //affect
  if(op>=50 && op<60) return 1; //unary
  if(op>=60 && op<100) return 2; //binary
  return 0; //hum
}

scr_val_t scr_val_t::oper(pvar_t *lv,int op,scr_val_t v0)
{
  scr_val_t r;
  switch(op) {
  case OPUM:
    r.type=v0.type;
    if(r.type==TYPEINT) r.v.i=-v0.v.i;
    if(r.type==TYPEDOUBLE) r.v.d=-v0.v.d;
    return r;
    break;
  case OPUN: 
    r.type=v0.type;
    if(r.type==TYPEINT) r.v.i=!v0.v.i;
    if(r.type==TYPEDOUBLE) r.v.d=!v0.v.d;
    return r;
    break;
  default:
    PRINTD("error in scr_val_t oper unary: default case");
  }
  return r;
}

scr_val_t scr_val_t::oper(pvar_t *lv,int op,scr_val_t v0,scr_val_t v1)
{
  scr_val_t r;
  // Affectations
  switch(op) {
  case OPAFF:
    //v0.print();
    //v1.print();
    if(v0.name==NULL) PRINTD("cannot asign");
    else {
      v1.transtype(v0.type);
      scr_setvar(lv,v0.name,v1);
      switch(v0.type) {
      case TYPEINT: v0.v.i=v1.v.i; break;
      case TYPEDOUBLE: v0.v.d=v1.v.d; break;
      case TYPESTRING: v0.v.s=strdup(v1.v.s); break;
      }
    }
    return v0;
    break;
  }

  switch(op) {
  case OPL: 
  case OPG: 
  case OPLE: 
  case OPGE:
  case OPEQ:
  case OPNEQ:

  case OPMUL:
  case OPDIV:
  case OPADD:
  case OPSUB:
    if(v0.type==TYPESTRING || v1.type==TYPESTRING) {
      // STRING
      if(v0.type!=TYPESTRING) v0.transtype(TYPESTRING);
      if(v1.type!=TYPESTRING) v1.transtype(TYPESTRING);
      if(op==OPADD) {
	//concat
	asprintf(&r.v.s,"%s%s",v0.v.s,v1.v.s);
	r.type=TYPESTRING;
	return r;
      }
      
      PRINTD("error in scr_val_t STRING");
      return v0;
    }

    if(v0.type==TYPEDOUBLE && v1.type==TYPEINT)
      v1.transtype(TYPEDOUBLE);
    if(v1.type==TYPEDOUBLE && v0.type==TYPEINT)
      v0.transtype(TYPEDOUBLE);
    if(v0.type==TYPEDOUBLE) {
      switch(op) {
      case OPL: r.type=TYPEINT; r.v.i=(v0.v.d < v1.v.d); break;
      case OPG: r.type=TYPEINT; r.v.i=(v0.v.d > v1.v.d); break;
      case OPLE: r.type=TYPEINT; r.v.i=(v0.v.d <= v1.v.d); break;
      case OPGE: r.type=TYPEINT; r.v.i=(v0.v.d >= v1.v.d); break;
      case OPEQ: r.type=TYPEINT; r.v.i=(v0.v.d == v1.v.d); break;
      case OPNEQ: r.type=TYPEINT; r.v.i=(v0.v.d != v1.v.d); break;
	
      case OPMUL: r.type=TYPEDOUBLE; r.v.d=(v0.v.d * v1.v.d); break;
      case OPDIV: r.type=TYPEDOUBLE; r.v.d=(v0.v.d / v1.v.d); break;
      case OPADD: r.type=TYPEDOUBLE; r.v.d=(v0.v.d + v1.v.d); break;
      case OPSUB: r.type=TYPEDOUBLE; r.v.d=(v0.v.d - v1.v.d); break;
      }
      return r;
    } else { //TYPEINT
      switch(op) {
      case OPL: r.type=TYPEINT; r.v.i=(v0.v.i < v1.v.i); break;
      case OPG: r.type=TYPEINT; r.v.i=(v0.v.i > v1.v.i); break;
      case OPLE: r.type=TYPEINT; r.v.i=(v0.v.i <= v1.v.i); break;
      case OPGE: r.type=TYPEINT; r.v.i=(v0.v.i >= v1.v.i); break;
      case OPEQ: r.type=TYPEINT; r.v.i=(v0.v.i == v1.v.i); break;
      case OPNEQ: r.type=TYPEINT; r.v.i=(v0.v.i != v1.v.i); break;

      case OPMUL: r.type=TYPEINT; r.v.i=(v0.v.i * v1.v.i); break;
      case OPDIV: r.type=TYPEINT; r.v.i=(v0.v.i / v1.v.i); break;
      case OPADD: r.type=TYPEINT; r.v.i=(v0.v.i + v1.v.i); break;
      case OPSUB: r.type=TYPEINT; r.v.i=(v0.v.i - v1.v.i); break;
      }
      return r;
    }
    break;
    
  case OPMOD:  
  case OPAND:  
  case OPOR: 
    if(v0.type==TYPEDOUBLE) v0.transtype(TYPEINT);
    if(v1.type==TYPEDOUBLE) v1.transtype(TYPEINT);
    r.type=TYPEINT;
    switch(op) {
    case OPAND:
      r.v.i=(v0.v.i && v1.v.i);
      break;
    case OPOR: 
      r.v.i=(v0.v.i || v1.v.i);
      break;
    case OPMOD:  
      r.v.i=(v0.v.i % v1.v.i);
      break;
    }
    return r;
    break;
    
  default:
    PRINTD("error in scr_val_t oper binary: default case");
  }
  
  return r;
}


void scr_val_t::print()
{
  switch(type) {
  case TYPEINT:
    printf("%d",v.i);
    break;
  case TYPEDOUBLE:
    printf("%f",v.d);
    break;
  case TYPESTRING:
    printf("%s",v.s);
    break;
  }
}

// FOR DEBUG

void scr_val_t::printoper(int op)
{
  switch(op) {
  case NOOP: printf("nop");break;
  case OPVAR: printf("var");break;
  case OPFCT: printf("fct");break;
  case OPCONF: printf("consti");break;
  case OPCONI: printf("constf");break;
  case OPCONS: printf("consts");break;
  case OPCAST: printf("cast");break;
  case OPAFF: printf("=");break;
  case OPUM: printf("-1");break;
  case OPUN: printf("!");break;
  case OPMUL: printf("*");break;
  case OPDIV: printf("/");break;
  case OPMOD: printf("%%");break;
  case OPADD: printf("+");break;
  case OPSUB: printf("-");break;
  case OPL: printf("<");break;
  case OPG: printf(">");break;
  case OPLE: printf("<=");break;
  case OPGE: printf(">=");break;
  case OPEQ: printf("==");break;
  case OPNEQ: printf("!=");break;
  case OPAND: printf("&&");break;
  case OPOR: printf("||");break;
  default: printf("undef");break;
  }
}
